import asyncio
import logging
from typing import Callable, Union

from django.apps import apps
from django.core.exceptions import ImproperlyConfigured

from django_guid.context import guid
from django_guid.utils import get_id_from_header, ignored_url

try:
    from django.utils.decorators import sync_and_async_middleware
except ImportError:  # pragma: no cover
    raise ImproperlyConfigured('Please use Django GUID 2.x for Django>=3.1. (`pip install django-guid>3`).')

from typing import TYPE_CHECKING

from django_guid.config import settings

if TYPE_CHECKING:
    from django.http import HttpRequest, HttpResponse

logger = logging.getLogger('django_guid')


def process_incoming_request(request: 'HttpRequest') -> None:
    """
    Processes an incoming request. This function is called before the view and later middleware.
    Same logic for both async and sync views.
    """
    if not ignored_url(request=request):
        # Process request and store the GUID in a contextvar
        guid.set(get_id_from_header(request))

        # Run all integrations
        for integration in settings.integrations:
            logger.debug('Running integration: `%s`', integration.identifier)
            integration.run(guid=guid.get())


def process_outgoing_request(response: 'HttpResponse', request: 'HttpRequest') -> None:
    """
    Process an outgoing request. This function is called after the view and before later middleware.
    """
    if not ignored_url(request=request):
        if settings.return_header:
            response[settings.guid_header_name] = guid.get()  # Adds the GUID to the response header
            if settings.expose_header:
                response['Access-Control-Expose-Headers'] = settings.guid_header_name

        # Run tear down for all the integrations
        for integration in settings.integrations:
            logger.debug('Running tear down for integration: `%s`', integration.identifier)
            integration.cleanup()


@sync_and_async_middleware
def guid_middleware(get_response: Callable) -> Callable:
    """
    Add this middleware to the top of your middlewares.
    """
    # One-time configuration and initialization.
    if not apps.is_installed('django_guid'):
        raise ImproperlyConfigured('django_guid must be in installed apps')
    # fmt: off
    if asyncio.iscoroutinefunction(get_response):
        async def middleware(request: 'HttpRequest') -> Union['HttpRequest', 'HttpResponse']:
            logger.debug('async middleware called')
            process_incoming_request(request=request)
            # ^ Code above this line is executed before the view and later middleware
            response = await get_response(request)
            process_outgoing_request(response=response, request=request)
            return response
    else:
        def middleware(request: 'HttpRequest') -> Union['HttpRequest', 'HttpResponse']:  # type: ignore
            logger.debug('sync middleware called')
            process_incoming_request(request=request)
            # ^ Code above this line is executed before the view and later middleware
            response = get_response(request)
            process_outgoing_request(response=response, request=request)
            return response
    # fmt: on
    return middleware



import logging
import uuid
from typing import TYPE_CHECKING, Optional, Union

from django_guid.config import settings

if TYPE_CHECKING:
    from django.http import HttpRequest, HttpResponse


logger = logging.getLogger('django_guid')


def get_correlation_id_from_header(request: 'HttpRequest') -> str:
    """
    Returns either the provided GUID or a new one depending on if the provided GUID is valid or not.
    :param request: HttpRequest object
    :return: GUID
    """
    given_guid: str = str(request.headers.get(settings.guid_header_name))
    if not settings.validate_guid:
        logger.debug('Returning ID from header without validating it as a GUID')
        return given_guid
    elif validate_guid(given_guid):
        logger.debug('%s is a valid GUID', given_guid)
        return given_guid
    else:
        new_guid = generate_guid()
        if all(letter.isalnum() or letter == '-' for letter in given_guid):
            logger.warning('%s is not a valid GUID. New GUID is %s', given_guid, new_guid)
        else:
            logger.warning('Non-alnum %s provided. New GUID is %s', settings.guid_header_name, new_guid)
        return new_guid


def get_id_from_header(request: 'HttpRequest') -> str:
    """
    Checks if the request contains the header specified in the Django settings.
    If it does, we fetch the header and attempt to validate the contents as GUID.
    If no header is found, we generate a GUID to be injected instead.
    :param request: HttpRequest object
    :return: GUID
    """
    header: str = request.headers.get(settings.guid_header_name)  # Case insensitive headers.get added in Django2.2
    if header:
        logger.info('%s found in the header', settings.guid_header_name)
        request.correlation_id = get_correlation_id_from_header(request)
    else:
        request.correlation_id = generate_guid()
        logger.info(
            'Header `%s` was not found in the incoming request. Generated new GUID: %s',
            settings.guid_header_name,
            request.correlation_id,
        )
    return request.correlation_id


def ignored_url(request: Union['HttpRequest', 'HttpResponse']) -> bool:
    """
    Checks if the current URL is defined in the `IGNORE_URLS` setting.

    :return: Boolean
    """
    is_ignored_url = request.get_full_path().strip("/") in settings.ignore_urls
    is_static_path = settings.static_url in request.path
    is_media_path = settings.media_url in request.path
    return any([is_ignored_url, is_static_path, is_media_path])


def generate_guid(uuid_length: Optional[int] = None) -> str:
    """
    Generates an UUIDv4/GUID as a string.

    :return: GUID
    """
    if settings.uuid_format == 'string':
        guid = str(uuid.uuid4())
    else:
        guid = uuid.uuid4().hex

    if uuid_length is None:
        return guid[: settings.uuid_length]
    return guid[:uuid_length]


def validate_guid(original_guid: str) -> bool:
    """
    Validates a GUID.

    :param original_guid: string to validate
    :return: bool
    """
    try:
        return bool(uuid.UUID(original_guid, version=4).hex)
    except ValueError:
        return False



from logging import Filter
from typing import TYPE_CHECKING

from django_guid.integrations.celery.context import celery_current, celery_parent

if TYPE_CHECKING:
    from logging import LogRecord


class CeleryTracing(Filter):
    # noinspection PyTypeHints
    def filter(self, record: 'LogRecord') -> bool:
        """
        Sets two record attributes: celery parent and celery current.
        Celery origin is the tracing ID of the process that spawned the current
        process, and celery current is the current process' tracing ID.

        In other words, if a worker sent a task to be executed by the worker pool,
        that celery worker's `current` tracing ID would become the next worker's `origin` tracing ID.
        """
        record.celery_parent_id: str = celery_parent.get()  # type: ignore
        record.celery_current_id: str = celery_current.get()  # type: ignore
        return True



from typing import Any

from celery.signals import setup_logging


@setup_logging.connect
def config_loggers(*args: Any, **kwargs: Any) -> None:  # pragma: no cover
    """
    Configures celery to use the Django settings.py logging configuration.
    """
    from logging.config import dictConfig

    from django.conf import settings

    dictConfig(settings.LOGGING)



import logging
from typing import Any

from django_guid.integrations import Integration

logger = logging.getLogger('django_guid')


class CeleryIntegration(Integration):
    """
    Passes correlation IDs from parent processes to child processes in a Celery context.

    This means a correlation ID can be transferred from a request to a worker, or from a worker to another worker.

    For workers executing scheduled tasks, a correlation ID is generated for each new task.
    """

    identifier = 'CeleryIntegration'

    def __init__(
        self,
        use_django_logging: bool = False,
        log_parent: bool = False,
        uuid_length: int = 32,
        sentry_integration: bool = False,
    ) -> None:
        """
        :param use_django_logging: If true, configures Celery to use the logging settings defined in settings.py
        :param log_parent: If true, traces the origin of a task. Should be True if you wish to use the CeleryTracing log filter.
        :param uuid_length: Optionally lets you set the length of the celery IDs generated for the log filter
        """
        super().__init__()
        self.log_parent = log_parent
        self.use_django_logging = use_django_logging
        self.uuid_length = uuid_length
        self.sentry_integration = sentry_integration

    def setup(self) -> None:
        """
        Loads Celery signals.
        """
        # Import pre-configured Celery signals that will pass on the correlation ID to a celery worker
        # or will generate a correlation ID when a worker starts a scheduled task
        from django_guid.integrations.celery.signals import before_task_publish, task_postrun, task_prerun  # noqa

        if self.use_django_logging:
            # Import pre-configured Celery signals that makes Celery adopt the settings.py log config
            from django_guid.integrations.celery.logging import config_loggers  # noqa

    def run(self, guid: str, **kwargs: Any) -> None:
        """
        Does nothing, as all we need for Celery tracing is to register signals during setup.
        """
        pass  # pragma: no cover



import logging
from typing import Any, Optional

from django.core.signals import request_finished
from django.dispatch import receiver

from django_guid.middleware import guid

logger = logging.getLogger('django_guid')


@receiver(request_finished)
def clear_guid(sender: Optional[dict], **kwargs: Any) -> None:
    """
    Receiver function for when a request finishes.

    When a request is finished, clear the GUID from the contextvar. This ensures a GUID is never passed down to
    the next request in sync views.

    :param sender: The sender of the signal. By documentation, we must allow this input parameter.
    :param kwargs: The request_finished signal does not actually send any kwargs, but Django will throw an error
        if we don't accept them. This is because at any point arguments could get added to the signal, and the receiver
        must be able to handle those new arguments.
    :return: None
    """
    logger.debug('Received signal `request_finished`, clearing guid')
    guid.set(None)



import logging

from django_guid.context import guid

logger = logging.getLogger('django_guid')


def get_guid() -> str:
    """
    Fetches the GUID of the current request
    """
    return guid.get()


def set_guid(new_guid: str) -> str:
    """
    Assigns a GUID to the current request
    """
    old_guid = guid.get()
    if old_guid:
        logger.info('Changing the guid ContextVar from %s to %s', old_guid, new_guid)
    guid.set(new_guid)
    return new_guid


def clear_guid() -> None:
    """
    Clears the GUID of the current request
    """
    old_guid = guid.get()
    if old_guid:
        logger.info('Clearing %s from the guid ContextVar', old_guid)
    guid.set(None)



from django_guid.integrations.base import Integration
from django_guid.integrations.celery import CeleryIntegration
from django_guid.integrations.sentry import SentryIntegration

__all__ = ['Integration', 'CeleryIntegration', 'SentryIntegration']



# flake8: noqa: D102
from typing import TYPE_CHECKING

from django.core.exceptions import ImproperlyConfigured

from django_guid.integrations import SentryIntegration

if TYPE_CHECKING:
    from django_guid.integrations.celery import CeleryIntegration  # pragma: no cover


class CeleryIntegrationSettings:
    def __init__(self, instance: 'CeleryIntegration') -> None:
        self.instance = instance
        self.validate()

    @property
    def use_django_logging(self) -> bool:
        return self.instance.use_django_logging

    @property
    def log_parent(self) -> bool:
        return self.instance.log_parent

    @property
    def uuid_length(self) -> int:
        return self.instance.uuid_length

    @property
    def sentry_integration(self) -> bool:
        return self.instance.sentry_integration

    def validate(self) -> None:
        if not isinstance(self.use_django_logging, bool):
            raise ImproperlyConfigured('The CeleryIntegration use_django_logging setting must be a boolean.')
        if not isinstance(self.log_parent, bool):
            raise ImproperlyConfigured('The CeleryIntegration log_parent setting must be a boolean.')
        if type(self.uuid_length) is not int or not 1 <= self.uuid_length <= 32:
            raise ImproperlyConfigured('The CeleryIntegration uuid_length setting must be an integer.')
        if not isinstance(self.sentry_integration, bool):
            raise ImproperlyConfigured('The CeleryIntegration sentry_integration setting must be a boolean.')
        if self.sentry_integration:
            SentryIntegration().setup()



import logging
from typing import TYPE_CHECKING, Any

from celery.signals import before_task_publish, task_postrun, task_prerun

from django_guid import clear_guid, get_guid, set_guid
from django_guid.config import settings
from django_guid.integrations.celery.context import celery_current, celery_parent
from django_guid.utils import generate_guid

if TYPE_CHECKING:
    from celery import Task

logger = logging.getLogger('django_guid.celery')

parent_header = 'CELERY_PARENT_ID'


def set_transaction_id(guid: str) -> None:
    """
    Sets the Sentry transaction ID if the Celery sentry integration setting is True.
    """
    if settings.integration_settings.celery.sentry_integration:
        from sentry_sdk import configure_scope

        with configure_scope() as scope:
            logger.debug('Setting Sentry transaction_id to %s', guid)
            scope.set_tag('transaction_id', guid)


@before_task_publish.connect
def publish_task_from_worker_or_request(headers: dict, **kwargs: Any) -> None:
    """
    Called when a request or celery worker publishes a task to the worker pool
    by calling task.delay(), task.apply_async() or using another equivalent method.
    This is where we transfer state from a parent process to a child process.
    """
    guid = get_guid()
    logger.info('Setting task request header as %s', guid)
    headers[settings.guid_header_name] = guid

    if settings.integration_settings.celery.log_parent:
        current = celery_current.get()
        if current:
            headers[parent_header] = current


@task_prerun.connect
def worker_prerun(task: 'Task', **kwargs: Any) -> None:
    """
    Called before a worker starts executing a task.
    Here we make sure to set the appropriate correlation ID for all logs logged
    during the tasks, and on the thread in general. In that regard, this does
    the Celery equivalent to what the django-guid middleware does for a request.
    """
    guid = task.request.get(settings.guid_header_name)
    if guid:
        logger.info('Setting GUID %s', guid)
        set_guid(guid)
        set_transaction_id(guid)
    else:
        generated_guid = generate_guid(uuid_length=settings.integration_settings.celery.uuid_length)
        logger.info('Generated GUID %s', generated_guid)
        set_guid(generated_guid)
        set_transaction_id(generated_guid)

    if settings.integration_settings.celery.log_parent:
        origin = task.request.get(parent_header)
        if origin:
            logger.info('Setting parent ID %s', origin)
            celery_parent.set(origin)
        generated_current_guid = generate_guid(uuid_length=settings.integration_settings.celery.uuid_length)
        logger.info('Generated current ID %s', generated_current_guid)
        celery_current.set(generated_current_guid)


@task_postrun.connect
def clean_up(task: 'Task', **kwargs: Any) -> None:
    """
    Called after a task is finished.
    Here we make sure to clean up the IDs we set in the pre-run method, so that
    the next task executed by the same worker doesn't inherit the same IDs.
    """
    logger.debug('Cleaning up GUIDs')
    clear_guid()

    if settings.integration_settings.celery.log_parent:
        celery_current.set(None)
        celery_parent.set(None)



# flake8: noqa: D102
from collections import defaultdict
from typing import Dict, List, Union

from django.conf import settings as django_settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.inspect import func_accepts_kwargs

from django_guid.integrations.celery.config import CeleryIntegrationSettings


class IntegrationSettings:
    def __init__(self, integration_settings: dict) -> None:
        self.settings = integration_settings

    @property
    def celery(self) -> CeleryIntegrationSettings:
        return CeleryIntegrationSettings(self.settings['CeleryIntegration'])

    def validate(self) -> None:
        if 'CeleryIntegration' in self.settings:
            self.celery.validate()


class Settings:
    def __init__(self) -> None:
        if hasattr(django_settings, 'DJANGO_GUID'):
            self.settings = django_settings.DJANGO_GUID
        else:
            self.settings = {}

    @property
    def guid_header_name(self) -> str:
        return self.settings.get('GUID_HEADER_NAME', 'Correlation-ID')

    @property
    def return_header(self) -> bool:
        return self.settings.get('RETURN_HEADER', True)

    @property
    def expose_header(self) -> bool:
        return self.settings.get('EXPOSE_HEADER', True)

    @property
    def ignore_urls(self) -> List[str]:
        return list({url.strip('/') for url in self.settings.get('IGNORE_URLS', [])})

    @property
    def validate_guid(self) -> bool:
        return self.settings.get('VALIDATE_GUID', True)

    @property
    def integrations(self) -> Union[list, tuple]:
        return self.settings.get('INTEGRATIONS', [])

    @property
    def integration_settings(self) -> IntegrationSettings:
        return IntegrationSettings({integration.identifier: integration for integration in self.integrations})

    @property
    def uuid_length(self) -> int:
        default_length: Dict[str, int] = defaultdict(lambda: 32, string=36)
        return self.settings.get('UUID_LENGTH', default_length[self.uuid_format])

    @property
    def uuid_format(self) -> str:
        return self.settings.get('UUID_FORMAT', 'hex')

    @property
    def static_url(self) -> str:
        return self.settings.get("STATIC_URL", "/static/")

    @property
    def media_url(self) -> str:
        return self.settings.get("MEDIA_URL", "/media/")

    def validate(self) -> None:
        if not isinstance(self.validate_guid, bool):
            raise ImproperlyConfigured('VALIDATE_GUID must be a boolean')
        if not isinstance(self.guid_header_name, str):
            raise ImproperlyConfigured('GUID_HEADER_NAME must be a string')  # Note: Case insensitive
        if not isinstance(self.return_header, bool):
            raise ImproperlyConfigured('RETURN_HEADER must be a boolean')
        if not isinstance(self.expose_header, bool):
            raise ImproperlyConfigured('EXPOSE_HEADER must be a boolean')
        if not isinstance(self.integrations, (list, tuple)):
            raise ImproperlyConfigured('INTEGRATIONS must be an array')
        if not isinstance(self.settings.get('IGNORE_URLS', []), (list, tuple)):
            raise ImproperlyConfigured('IGNORE_URLS must be an array')
        if not all(isinstance(url, str) for url in self.settings.get('IGNORE_URLS', [])):
            raise ImproperlyConfigured('IGNORE_URLS must be an array of strings')
        if type(self.uuid_length) is not int or self.uuid_length < 1:
            raise ImproperlyConfigured('UUID_LENGTH must be an integer and positive')
        if self.uuid_format == 'string' and not 1 <= self.uuid_length <= 36:
            raise ImproperlyConfigured('UUID_LENGTH must be between 1-36 when UUID_FORMAT is string')
        if self.uuid_format == 'hex' and not 1 <= self.uuid_length <= 32:
            raise ImproperlyConfigured('UUID_LENGTH must be between 1-32 when UUID_FORMAT is hex')
        if self.uuid_format not in ('hex', 'string'):
            raise ImproperlyConfigured('UUID_FORMAT must be either hex or string')

        self._validate_and_setup_integrations()

    def _validate_and_setup_integrations(self) -> None:
        """
        Validate the INTEGRATIONS settings and verify each integration
        """
        self.integration_settings.validate()
        for integration in self.integrations:
            # Make sure all integration methods are callable
            for method, name in [
                (integration.setup, 'setup'),
                (integration.run, 'run'),
                (integration.cleanup, 'cleanup'),
            ]:
                # Make sure the methods are callable
                if not callable(method):
                    raise ImproperlyConfigured(
                        f'Integration method `{name}` needs to be made callable for `{integration.identifier}`'
                    )

                # Make sure the method takes kwargs
                if name in ['run', 'cleanup'] and not func_accepts_kwargs(method):
                    raise ImproperlyConfigured(
                        f'Integration method `{name}` must '
                        f'accept keyword arguments (**kwargs) for `{integration.identifier}`'
                    )

            # Run validate method
            integration.setup()


settings = Settings()



import django

from django_guid.api import clear_guid, get_guid, set_guid  # noqa F401

__version__ = '3.4.0'

if django.VERSION < (3, 2):
    default_app_config = 'django_guid.apps.DjangoGuidConfig'

__all__ = ['clear_guid', 'get_guid', 'set_guid']
